@@ -740,17 +740,16 @@ nsContentUtils::InProlog(nsIDOMNode *aNode)
 // static
 nsresult
 nsContentUtils::doReparentContentWrapper(nsIContent *aChild,
-                                         nsIDocument *aNewDocument,
-                                         nsIDocument *aOldDocument,
                                          JSContext *cx,
+                                         JSObject *aOldGlobal,
                                          JSObject *parent_obj)
 {
   nsCOMPtr<nsIXPConnectJSObjectHolder> old_wrapper;
 
   nsresult rv;
 
-  rv = sXPConnect->ReparentWrappedNativeIfFound(cx, ::JS_GetGlobalObject(cx),
-                                                parent_obj, aChild,
+  rv = sXPConnect->ReparentWrappedNativeIfFound(cx, aOldGlobal, parent_obj,
+                                                aChild,
                                                 getter_AddRefs(old_wrapper));
   NS_ENSURE_SUCCESS(rv, rv);
 
@@ -771,7 +770,7 @@ nsContentUtils::doReparentContentWrapper(nsIContent *aChild,
     nsIContent *child = aChild->GetChildAt(i);
     NS_ENSURE_TRUE(child, NS_ERROR_UNEXPECTED);
 
-    rv = doReparentContentWrapper(child, aNewDocument, aOldDocument, cx, old);
+    rv = doReparentContentWrapper(child, cx, aOldGlobal, old);
     NS_ENSURE_SUCCESS(rv, rv);
   }
 
@@ -779,7 +778,7 @@ nsContentUtils::doReparentContentWrapper(nsIContent *aChild,
 }
 
 static JSContext *
-GetContextFromDocument(nsIDocument *aDocument)
+GetContextFromDocument(nsIDocument *aDocument, JSObject** aGlobalObject)
 {
   nsIScriptGlobalObject *sgo = aDocument->GetScriptGlobalObject();
 
@@ -789,6 +788,8 @@ GetContextFromDocument(nsIDocument *aDocument)
     return nsnull;
   }
 
+  *aGlobalObject = sgo->GetGlobalJSObject();
+  
   nsIScriptContext *scx = sgo->GetContext();
 
   if (!scx) {
@@ -823,23 +824,28 @@ nsContentUtils::ReparentContentWrapper(nsIContent *aContent,
   nsISupports* new_parent = aNewParent ? (nsISupports*)aNewParent :
     (nsISupports*)aNewDocument;
 
-  JSContext *cx = GetContextFromDocument(aOldDocument);
+  // Make sure to get our hands on the right scope object, since
+  // GetWrappedNativeOfNativeObject doesn't call PreCreate and hence won't get
+  // the right scope if we pass in something bogus.  The right scope lives on
+  // the script global of the old document.
+  // XXXbz note that if GetWrappedNativeOfNativeObject did call PreCreate it
+  // would get the wrong scope (that of the _new_ document), so we should be
+  // glad it doesn't!
+  JSObject *globalObj;
+  JSContext *cx = GetContextFromDocument(aOldDocument, &globalObj);
 
-  if (!cx) {
-    // No JSContext left in the old scope, can't find the old wrapper
-    // w/o the old context.
+  if (!cx || !globalObj) {
+    // No JSContext left in the old scope, or no global object around; can't
+    // find the old wrapper w/o the old context or global object
 
     return NS_OK;
   }
 
   nsCOMPtr<nsIXPConnectWrappedNative> wrapper;
-
-  nsresult rv;
-
-  rv = sXPConnect->GetWrappedNativeOfNativeObject(cx, ::JS_GetGlobalObject(cx),
-                                                  aContent,
-                                                  NS_GET_IID(nsISupports),
-                                                  getter_AddRefs(wrapper));
+  nsresult rv =
+    sXPConnect->GetWrappedNativeOfNativeObject(cx, globalObj, aContent,
+                                               NS_GET_IID(nsISupports),
+                                               getter_AddRefs(wrapper));
   NS_ENSURE_SUCCESS(rv, rv);
 
   if (!wrapper) {
@@ -849,8 +855,9 @@ nsContentUtils::ReparentContentWrapper(nsIContent *aContent,
     return NS_OK;
   }
 
-  // Wrap the new parent and reparent aContent
-
+  // Wrap the new parent and reparent aContent.  Don't bother using globalObj
+  // here, since it's wrong for the new parent anyway...  Luckily, WrapNative
+  // will PreCreate and hence get the right scope.
   nsCOMPtr<nsIXPConnectJSObjectHolder> holder;
   rv = sXPConnect->WrapNative(cx, ::JS_GetGlobalObject(cx), new_parent,
                               NS_GET_IID(nsISupports),
@@ -861,8 +868,7 @@ nsContentUtils::ReparentContentWrapper(nsIContent *aContent,
   rv = holder->GetJSObject(&obj);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  return doReparentContentWrapper(aContent, aNewDocument, aOldDocument, cx,
-                                  obj);
+  return doReparentContentWrapper(aContent, cx, globalObj, obj);
 }
 
 nsIDocShell *
@@ -924,6 +930,37 @@ nsContentUtils::IsCallerChrome()
   return is_caller_chrome;
 }
 
+static PRBool IsCallerTrustedForCapability(const char* aCapability)
+{
+  if (nsContentUtils::IsCallerChrome())
+    return PR_TRUE;
+
+  // The secman really should handle UniversalXPConnect case, since that
+  // should include UniversalBrowserRead... doesn't right now, though.
+  PRBool hasCap;
+  nsIScriptSecurityManager* ssm = nsContentUtils::GetSecurityManager();
+  if (NS_FAILED(ssm->IsCapabilityEnabled(aCapability, &hasCap)))
+    return PR_FALSE;
+  if (hasCap)
+    return PR_TRUE;
+    
+  if (NS_FAILED(ssm->IsCapabilityEnabled("UniversalXPConnect", &hasCap)))
+    return PR_FALSE;
+  return hasCap;
+}
+
+PRBool
+nsContentUtils::IsCallerTrustedForRead()
+{
+  return IsCallerTrustedForCapability("UniversalBrowserRead");
+}
+
+PRBool
+nsContentUtils::IsCallerTrustedForWrite()
+{
+  return IsCallerTrustedForCapability("UniversalBrowserWrite");
+}
+
 // static
 PRBool
 nsContentUtils::InSameDoc(nsIDOMNode* aNode, nsIDOMNode* aOther)
@@ -2352,7 +2389,14 @@ nsContentUtils::NotifyXPCIfExceptionPending(JSContext* aCx)
   nsCOMPtr<nsIXPCNativeCallContext> nccx;
   XPConnect()->GetCurrentNativeCallContext(getter_AddRefs(nccx));
   if (nccx) {
-    nccx->SetExceptionWasThrown(PR_TRUE);
+    // Check to make sure that the JSContext that nccx will mess with is the
+    // same as the JSContext we've set an exception on.  If they're not the
+    // same, don't mess with nccx.
+    JSContext* cx;
+    nccx->GetJSContext(&cx);
+    if (cx == aCx) {
+      nccx->SetExceptionWasThrown(PR_TRUE);
+    }
   }
 }
 
